import string
from mssqldb import *

__version__ = '0.04'

# DB-API values
apilevel = '2.0'                        # DB API level supported

threadsafety = 3                        # Threads may share the
                                        # module, connections and
                                        # cursors.

paramstyle = 'qmark'                    # Question mark style,
                                        # e.g. '...WHERE name=?'

# DB-API exceptions
#
# StandardError
# |__Warning
# |__Error
#    |__InterfaceError
#    |__DatabaseError
#       |__DataError
#       |__OperationalError
#       |__IntegrityError
#       |__InternalError
#       |__ProgrammingError
#       |__NotSupportedError

class Warning(StandardError):
    pass

class Error(StandardError):
    pass

class InterfaceError(Error):
    pass

class DatabaseError(Error):
    pass

class DataError(DatabaseError):
    pass

class OperationalError(DatabaseError):
    pass

class IntegrityError(DatabaseError):
    pass

class InternalError(DatabaseError):
    pass

class ProgrammingError(DatabaseError):
    pass

class NotSupportedError(DatabaseError):
    pass

_msgs = []

def err_handler(dbproc, severity, dberr, oserr, dberrstr, oserrstr):
    global _msgs
    parts = ['Error:']
    parts.append('  %s' % dberrstr)
    if oserrstr:
        parts.append('  oserr: %s' % oserrstr)
    if _msgs:
        for msg in _msgs:
            parts.append(str(msg))
    _msgs = []
    raise Error(string.join(parts, '\n'))

class Msg:
    def __init__(self, msgno, msgstate, severity, msgtext,
                 srvname, procname, line):
        self.msgno = msgno   
        self.msgstate = msgstate
        self.severity = severity
        self.msgtext = msgtext 
        self.srvname = srvname 
        self.procname = procname
        self.line = line

    def __str__(self):
        parts = []
        parts.append('message:')
        if self.msgno is not None:
            parts.append('  msgno:    %s' % self.msgno)
        if self.msgstate is not None:
            parts.append('  msgstate: %s' % self.msgstate)
        if self.severity is not None:
            parts.append('  severity: %s' % self.severity)
        if self.msgtext is not None:
            parts.append('  msgtext:  %s' % self.msgtext)
        if self.srvname:
            parts.append('  srvname:  %s' % self.srvname)
        if self.procname:
            parts.append('  procname: %s' % self.procname)
        if self.line is not None:
            parts.append('  line:     %s' % self.line)
        return string.join(parts, '\n')

def msg_handler(dbproc, msgno, msgstate, severity, msgtext,
                srvname, procname, line):
    global _msgs
    _msgs.append(Msg(msgno, msgstate, severity, msgtext,
                     srvname, procname, line))
    return 0

# Setup global library context
if not dbinit():
    raise InternalError('dbinit failed')
dberrhandle(err_handler)
dbmsghandle(msg_handler)

def _extract_row(dbproc, bufs):
    '''Extract a row tuple from buffers.
    '''
    row = [None] * len(bufs)
    for i in range(len(bufs)):
        buf = bufs[i]
        col_len = dbproc.dbdatlen(i + 1)
        if col_len:
            buf.len = col_len
            row[i] = buf.get()
        else:
            row[i] = None
    return tuple(row)

def _row_bind(dbproc):
    '''Bind buffers for column data.
    '''
    num_cols = dbproc.dbnumcols()
    bufs = []
    for i in range(1, num_cols + 1):
        buf = DataBuf(dbproc.dbcoltype(i), dbproc.dbcollen(i))
        dbproc.dbbind(i, buf, 0)
        bufs.append(buf)
    return bufs

def _extract_compute(dbproc, row_code):
    '''Extract a row tuple from buffers.
    '''
    num_cols = dbproc.dbnumalts(row_code)
    row = [None] * num_cols
    for i in range(num_cols):
        col_num = i + 1
        col_len = dbproc.dbaltlen(row_code, col_num)
        if col_len:
            buf = DataBuf(dbproc.dbalttype(row_code, col_num), col_len)
            dbproc.dbadata(buf, row_code, col_num)
            buf.len = col_len
            row[i] = buf.get()
        else:
            row[i] = None
    return tuple(row)

def _make_databuf(v):
    if type(v) is type(0):
        buf = DataBuf(SQLINT4)
    elif type(v) is type(''):
        buf = DataBuf(SQLCHAR, len(v))
    elif type(v) is type(0.0):
        buf = DataBuf(SQLFLT8)
    else:
        raise DataError('unhandled data type')
    buf.set(v)
    return buf

CMD_PREPARED = 0                        # prepared command
CMD_END_RESULTS = 1                     # all result sets fetched
CMD_FETCHING = 2                        # fetching rows
CMD_COMPUTE_WAITING = 3                 # fetching rows - next set is compute
CMD_COMPUTE = 4                         # next set is compute
CMD_FETCH_WAITING = 5                   # fetching compute - next set is rows
CMD_FETCH_FIRST = 6                     # fetching compute - next set is rows
CMD_END_SET = 7                         # current set finished

class Cursor:
    def __init__(self, owner):
        '''Implements DB-API Cursor object
        '''
        self.description = None         # DB-API
        self.rowcount = -1              # DB-API
        self.arraysize = 1              # DB-API
        self._owner = owner
        self._dbproc = None
        self._sql = None

    def __del__(self):
        self._cancel()

    def callproc(self, name, params = {}):
        '''DB-API Cursor.callproc()
        '''
        dbproc = self._dbproc = self._owner._new_dbproc()
        if dbproc.dbrpcinit(name) == FAIL:
            raise InternalError('dbrpcinit')
        bufs = []
        for k, v in params.items():
            buf = _make_databuf(v)
            bufs.append(buf)
            if dbproc.dbrpcparam(k, buf) == FAIL:
                raise InternalError('dbrpcparam')
        if dbproc.dbrpcsend() == FAIL:
            raise InternalError('dbrpcsend')
        if dbproc.dbsqlok() == FAIL:
            raise InternalError('dbsqlok')
        self._state = CMD_PREPARED

    def close(self):
        '''DB-API Cursor.close()
        '''
        self._cancel()

    def execute(self, sql, params = []):
        '''DB-API Cursor.execute()
        '''
        if self._sql:
            if self._sql != sql:
                self._cancel()
                self._sql = None
            else:
                self._discard_results()
        if not self._sql:
            self._prepare(sql)

    def executemany(self, sql, params_seq = []):
        '''DB-API Cursor.executemany()
        '''
        for params in params_seq:
            self.execute(sql, params)

    def fetchone(self):
        '''DB-API Cursor.fetchone()
        '''
        dbproc = self._dbproc
        if self._state == CMD_FETCHING:
            row_code = dbproc.dbnextrow()
            if row_code == NO_MORE_ROWS:
                self._state = CMD_END_SET
            elif row_code == REG_ROW:
                return _extract_row(dbproc, self._bufs)
            else:
                self._next_row = _extract_compute(dbproc, row_code)
                self._state = CMD_COMPUTE_WAITING
        elif self._state == CMD_COMPUTE:
            row = self._next_row
            row_code = dbproc.dbnextrow()
            if row_code == NO_MORE_ROWS:
                self._state = CMD_END_SET
            elif row_code == REG_ROW:
                self._next_row = _extract_row(dbproc, self._bufs)
                self._state = CMD_FETCH_WAITING
            else:
                self._next_row = _extract_compute(dbproc, row_code)
                self._state = CMD_COMPUTE_WAITING
            return row
        elif self._state == CMD_COMPUTE_WAITING:
            pass
        elif self._state == CMD_FETCH_WAITING:
            pass
        elif self._state == CMD_FETCH_FIRST:
            self._state = CMD_FETCHING
            return self._next_row

    def fetchmany(self, num = -1):
        '''DB-API Cursor.fetchmany()
        '''
        if num == -1:
            num = self.arraysize
        rows = []
        for i in xrange(num):
            row = self.fetchone()
            if not row:
                break
            rows.append(row)
        return rows

    def fetchall(self):
        '''DB-API Cursor.fetchall()
        '''
        rows = []
        while 1:
            row = self.fetchone()
            if not row:
                break
            rows.append(row)
        return rows

    def nextset(self):
        '''DB-API Cursor.nextset()
        '''
        if self._state == CMD_END_RESULTS:
            return
        if self._state in (CMD_COMPUTE, CMD_FETCH_FIRST):
            self.fetchall()
        if self._state == CMD_COMPUTE_WAITING:
            self._state = CMD_COMPUTE
            return 1
        if self._state == CMD_FETCH_WAITING:
            self._state = CMD_FETCH_FIRST
            return 1
        if self._state != CMD_END_SET:
            self._dbproc.dbcanquery()
        self._start_result_set()
        return self._state != CMD_END_RESULTS or None

    def setinputsizes(self):
        '''DB-API Cursor.setinputsizes()
        '''
        pass

    def setoutputsize(self, size, column = None):
        '''DB-API Cursor.setoutputsize()
        '''
        pass

    def _cancel(self):
        if self._dbproc:
            self._dbproc.dbcancel()

    def _discard_results(self):
        '''Discard all pending results and prepare to execute again.
        '''
        while self._state != CMD_END_RESULTS:
            self.nextset()

    def _prepare(self, sql):
        '''Prepare the statement to be executed for this cursor.
        '''
        dbproc = self._dbproc = self._owner._new_dbproc()
        # send command to server
        self._sql = sql
        dbproc.dbcmd(self._sql)
        dbproc.dbsqlexec()
        self._state = CMD_PREPARED
        self._start_result_set()

    def _start_result_set(self):
        dbproc = self._dbproc
        while 1:
            result_code = dbproc.dbresults()
            if result_code == NO_MORE_RESULTS:
                self._state = CMD_END_RESULTS
                return
            elif result_code != SUCCEED:
                raise InternalError('dbresults')
            bufs = self._bufs = _row_bind(dbproc)
            desc = []
            for i in range(len(bufs)):
                buf = bufs[i]
                desc.append((dbproc.dbcolname(i + 1), buf.type, 0,
                             buf.size, 0, 0, 0))
            self.description = desc
            self._state = CMD_FETCHING
            return

class Connection:
    def __init__(self, dsn, user, passwd, database = None, strip = 0):
        '''DB-API Sybase.Connect()
        '''
        self.dsn = dsn
        self.user = user
        self.passwd = passwd
        self.database = database
        self.login = dblogin()
        self.login.dbsetluser(user)
        self.login.dbsetlpwd(passwd)
        self.dbproc = self._new_dbproc()

    def _new_dbproc(self):
        dbproc = self.login.dbopen(self.dsn)
        if self.database:
            dbproc.dbuse(self.database)
        return dbproc

    def cursor(self):
        '''DB-API Connection.cursor()
        '''
        return Cursor(self)

    def execute(self, sql):
        '''Backwards compatibility
        '''
        self.dbproc.dbcmd(sql)
        self.dbproc.dbsqlexec()
        result_list = self._fetch_results()
        return result_list

    def _fetch_results(self):
        result_list = []
        dbproc = self.dbproc
        while 1:
            result_code = dbproc.dbresults()
            if result_code == NO_MORE_RESULTS:
                break
            if result_code == SUCCEED:
                row_bufs = _row_bind(dbproc)
		logical_result = []
                while 1:
                    row_code = dbproc.dbnextrow()
                    if row_code == NO_MORE_ROWS:
                        break
                    if row_code == REG_ROW:
                        row = _extract_row(dbproc, row_bufs)
                        if row:
                            logical_result.append(row)
                        else:
                            break
                    else:
                        if logical_result:
                            result_list.append(logical_result)
                            logical_result = []
                        row = _extract_compute(dbproc, row_code)
                        if row:
                            result_list.append([row])
                if logical_result:
                    result_list.append(logical_result)
        return result_list

def connect(dsn, user, passwd, database = None, strip = 0):
    return Connection(dsn, user, passwd, database, strip)
